# **********************************************************************
# Copyright 2017-2018 VMware, Inc. All rights reserved.
# -- VMware Confidential
# **********************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# **********************************************************************
import json
import nsx_logging
import os
import sys
from cli.utils.command_utils import invoke_command
from cli.utils.node_utils import get_node_type
from cli.utils.constants import NODE_TYPE_HOST_CENTOSKVM
from cli.utils.constants import NODE_TYPE_HOST_SLESKVM
from cli.utils.constants import NODE_TYPE_HOST_ESX
from cli.utils.constants import NODE_TYPE_HOST_RHELKVM
from cli.utils.constants import NODE_TYPE_HOST_UBUNTUKVM
from cli.utils.constants import NODE_TYPE_HOST_WIN
from cli.utils.constants import NODE_TYPE_BAREMETAL_UBUNTU_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_SLES_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_OEL_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_WINDOWS_SERVER
from cli.utils.constants import NODE_TYPE_BAREMETAL_RHEL_CONTAINER
from cli.utils.constants import NODE_TYPE_BAREMETAL_CENTOS_CONTAINER
from cli.utils.constants import OUTPUT_MODIFIER_COUNT
from cli.utils.constants import OUTPUT_MODIFIER_FIND
from cli.utils.constants import OUTPUT_MODIFIER_FIRST
from cli.utils.constants import OUTPUT_MODIFIER_IGNORE
from cli.utils.constants import OUTPUT_MODIFIER_JSON
from cli.utils.constants import OUTPUT_MODIFIER_LAST
from cli.utils.constants import OUTPUT_MODIFIER_SORT
from cli.utils.location_utils import FilePaths
from cli.utils.error_code import ErrorCode as ec
from cli.commands.cli_command_provider import CliCommandProvider
from cli.utils.render_utils import format_subprocess_output
from cli.commands.common.arguments import IntegerArgument

lg = nsx_logging.getLogger("cli.commands.host_shared.firewall")

#
# DFW related commands
#

_FIREWALL_CLI = {
   NODE_TYPE_HOST_SLESKVM: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_HOST_CENTOSKVM: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_HOST_ESX: {
      'bin': "/opt/vmware/nsx-cli/bin/nsx-appctl",
      'socket': "/var/run/vmware/nsx-cfgagent/cfgagent-cli"
   },
   NODE_TYPE_HOST_RHELKVM: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_HOST_UBUNTUKVM: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_HOST_WIN: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_BAREMETAL_UBUNTU_SERVER: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_BAREMETAL_RHEL_SERVER: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_BAREMETAL_CENTOS_SERVER: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_BAREMETAL_SLES_SERVER: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_BAREMETAL_OEL_SERVER: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_BAREMETAL_RHEL_CONTAINER: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_BAREMETAL_CENTOS_CONTAINER: {
      'bin': "/usr/bin/ovs-appctl",
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
   NODE_TYPE_BAREMETAL_WINDOWS_SERVER: {
      'bin': FilePaths.OVS_APPCTL_PATH,
      'socket': FilePaths.NSXA_CTL_SOCKET
   },
}


def has_firewall_cli(node_type):
   return (node_type == NODE_TYPE_HOST_SLESKVM or
           node_type == NODE_TYPE_HOST_CENTOSKVM or
           node_type == NODE_TYPE_HOST_ESX or
           node_type == NODE_TYPE_HOST_RHELKVM or
           node_type == NODE_TYPE_HOST_UBUNTUKVM or
           node_type == NODE_TYPE_BAREMETAL_UBUNTU_SERVER or
           node_type == NODE_TYPE_BAREMETAL_RHEL_SERVER or
           node_type == NODE_TYPE_BAREMETAL_CENTOS_SERVER or
           node_type == NODE_TYPE_BAREMETAL_OEL_SERVER or
           node_type == NODE_TYPE_BAREMETAL_SLES_SERVER or
           node_type == NODE_TYPE_BAREMETAL_WINDOWS_SERVER or
           node_type == NODE_TYPE_BAREMETAL_RHEL_CONTAINER or
           node_type == NODE_TYPE_BAREMETAL_CENTOS_CONTAINER or
           node_type == NODE_TYPE_HOST_WIN)


def invoke_firewall_command(cmd_list):
   node_type = get_node_type()
   if has_firewall_cli(node_type):
      cmd = [_FIREWALL_CLI[node_type]['bin'], "-t", _FIREWALL_CLI[node_type]['socket']]
      cmd.extend(cmd_list)
      rc, out, err = invoke_command(cmd)
      out2 = { "results": "%s" % out }
      return (rc, json.dumps(out2), err)
   else:
      lg.error("node type %s does not have firewall CLI", node_type,
               error_code=ec.INVALID_ARGUMENT)
      return (-1, None, "invalid command")


def get_firewall_cli_output(command):
   cmd_list = command
   rc, out, err = invoke_firewall_command(cmd_list)
   if rc != 0:
      lg.error("Failed to invoke firewall CLI command, rc: %d, err: %s", rc, err,
               error_code=ec.OPERATION_FAILED)
      out = None

   return out

# arguments
class DFWCountArgument(IntegerArgument):
    """Verify argument to validate line count value."""
    _MIN_VALUE = 0
    _MAX_VALUE = 80


class DFWVifArgument(object):
   """Query argument to help with auto-complete of vifs"""
   def __init__(self, descriptor):
      self._descriptor = descriptor

   def __call__(self, req_dict):
      res = get_firewall_cli_output(["dfw/vif", "-json"])
      if res is None:
         lg.error("error in getting dfw/vif")
         return json.dumps([])
      else:
         viflist = []
         resjson = json.loads(res)
         vifs = json.loads(resjson.get("results")).get("dfw_vif_list")
         for vif in vifs:
            if "vif_id" in vif:
               viflist.append(vif.get("vif_id"))
         return json.dumps(viflist)

# commands

class FirewallShowBase(CliCommandProvider):
   CMD_STR = None
   VIF_ARG = 0
   def __call__(self, cmd_list):
      cmd_result = ""
      subcmd_list = [self.CMD_STR]
      if self.VIF_ARG:
         subcmd_list.extend([cmd_list[2]])
      subcmd_list.extend(["-json"])
      res = get_firewall_cli_output(subcmd_list)
      if res is None:
         cmd_result = "MSG_FIREWALL_COMM_ERROR"
      else:
         resjson = json.loads(res)
         data = resjson.get("results")
         cmd_result = "OK\n\n%s" % data

      return cmd_result


class FirewallShowVifs(FirewallShowBase):
   CMD_STR = "dfw/vif"
   VIF_ARG = 0


class FirewallShowRulesetRules(FirewallShowBase):
   CMD_STR = "dfw/rules"
   VIF_ARG = 1


class FirewallShowAddrSets(FirewallShowBase):
   CMD_STR = "dfw/addrsets"
   VIF_ARG = 1


class FirewallShowExclusion(FirewallShowBase):
   CMD_STR = "dfw/exclude"
   VIF_ARG = 0


class FirewallShowStatus(FirewallShowBase):
   CMD_STR = "dfw/switch"
   VIF_ARG = 0


class FirewallShowProfiles(FirewallShowBase):
   CMD_STR = "dfw/profiles"
   VIF_ARG = 1


class FirewallGetLog(CliCommandProvider):
    def __call__(self, cmd_list):
        if sys.platform == "win32":
            from cli.utils.location_utils import FilePaths
            logfile_list = []
            for log in os.listdir(FilePaths.NSX_LOG_DIR):
                if log.startswith("Openflow_controller_cpp.log"):
                    logfile_list.append(log)
            logfile_list = sorted(logfile_list, key=lambda x: os.path.getmtime(os.path.join(FilePaths.NSX_LOG_DIR, x)))
            lines = []
            for log in logfile_list:
                if not os.path.isfile(os.path.join(FilePaths.NSX_LOG_DIR, log)):
                    continue
                with open(os.path.join(FilePaths.NSX_LOG_DIR, log), 'r') as f:
                    for line in f:
                        if "FIREWALL-PKTLOG" in line:
                            lines.append(line)
                f.close()
            if not lines:
                lines = ["Packet log file does not exist."]
        else:
            log_name = "/var/log/dfwpktlogs.log"
            lines = []
            lastcount = 0;
            if len(cmd_list) > 4:
                lastcount = int(cmd_list[4])
            if not os.path.exists(log_name):
                lines = ["Packet log file does not exist."]
            else:
                with open(log_name, 'r') as f:
                    lines = [line.rstrip() for line in f]
                f.close()
                if lastcount > 0:
                    if lastcount > len(lines):
                        lastcount = len(lines)
                    lines = lines[-lastcount:]
        return "OK\n\n%s" % json.dumps(lines)
