# ***************************************************************************
# Copyright 2015-2017 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************
import nsx_logging
import os
import socket
import ctypes
import sys
import subprocess
import re
from cli.utils.exceptions import IoctlResultException
from cli.utils.command_utils import invoke_command
from cli.utils.location_utils import FilePaths
from cli.utils.error_code import ErrorCode as ec

lg = nsx_logging.getLogger("cli.commands.host_shared.ovs_util")

def run_command(cmd, split = True, header = True):
    assert(isinstance(cmd, list))
    result = None
    rc, out, err = invoke_command(cmd)
    if rc != 0:
        lg.error("Error invoking cmd: %s, rc: %s, "
                 "out: %s, err: %s", cmd, rc, out, err,
                 error_code=ec.OPERATION_FAILED)
        return result
    if split:
        result = [[x.strip() for x in ln.split() if x] for ln in out.splitlines()]
    else:
        result = [x.strip() for x in out.splitlines()]
    if header and len(result) >= 1:
        del result[0]
    return result

def run_ovs_appctl(target, command='', args=[],
                   split = True, header = True):
    ovs_appctl_cmd = os.path.join(FilePaths.OVS_BIN_DIR, 'ovs-appctl')
    if isinstance(args, list):
        cmd_list = [ovs_appctl_cmd, '-t', target, command] + args
    else:
        cmd_list = [ovs_appctl_cmd, '-t', target, command, str(args)]
    return run_command(cmd_list, split, header)


def run_ovs_vsctl(command, args=[], fmts=[], returnDict=True):
    ovs_vsctl_cmd = os.path.join(FilePaths.OVS_BIN_DIR, 'ovs-vsctl')
    if isinstance(fmts, list):
        cmd_list = [ovs_vsctl_cmd] + fmts + [command]
    else:
        cmd_list = [ovs_vsctl_cmd, str(fmts), command]
    if isinstance(args, list):
        cmd_list = cmd_list + args
    else:
        cmd_list = cmd_list + [str(args)]
    result_list = run_command(cmd_list, True, False)
    # If do not need the result as a dict, return list here
    if not returnDict:
        return result_list
    # Create a dictionary from the result list.
    result_dict = {}
    for attr in result_list:
        key = attr[0]
        if key[len(key) - 1] == ':':
            key = key.rstrip(':')
            val = attr[1]
        else:
            val = attr[2]
        result_dict[key] = val
    return result_dict

def run_nsxa_ctl(command, args=[], split = True, header = True):
    try:
        return run_ovs_appctl(FilePaths.NSXA_CTL_SOCKET, command, args,
                              split, header)
    except:
        return None

def run_exporter_cli(command, args=[], split = True, header = True):
    return run_ovs_appctl(FilePaths.EXPORTER_CLI_SOCKET, command, args,
                          split, header)

def run_l3d_ctl(command, args=[], split = True, header = True):
    return run_ovs_appctl('ovs-l3d', command, args, split, header)

def table_to_map(table, key_column):
    map = {}
    for entry in table:
        key = entry[key_column]
        if key in table:
            map[key].append(entry)
        else:
            map[key] = [entry]
    return map

def remove_header(table):
    del table[0]
    return table
def remove_footer(table):
    if len(table) > 0:
        del table[-1]
    return table

def get_vni_list():
    return run_nsxa_ctl('vni/list')

def get_vlan_list():
    return run_nsxa_ctl('vlan/list')

def get_spoofguard_whitelist(lport):
    return run_nsxa_ctl('lport/white-list', lport)

def get_switchsecurity_config(lport):
    return run_nsxa_ctl('lport/switch-security-config', lport)

def get_spoofguard_stats(lport):
    stats = run_exporter_cli('lport/spoof-guard-stats', lport, split = False,
                             header = False)
    if len(stats) < 2:
        lg.error('Error retrieving spoofguard stats for lport %s' % lport)
        return None
    return stats

def get_switchsecurity_stats(lport):
    stats = run_exporter_cli('lport/switch-security-stats', lport,
                             split = False, header = False)
    if len(stats) < 2:
        lg.error('Error retrieving switch-security stats for lport %s' % lport)
        return None
    return stats

def get_lport_list(id = None):
    if id == None:
        return run_nsxa_ctl('lport/list')
    elif id[0]:
        return run_nsxa_ctl('lport/list', ['VNI', str(id[1])])
    else:
        return run_nsxa_ctl('lport/list', ['VLAN', str(id[1])])

def get_ip_discovery_config(lport):
    if lport is not None:
        return run_nsxa_ctl('lport/ip-discovery-profile', lport)
    else:
        return run_nsxa_ctl('lport/ip-discovery-profile')

def get_ip_discovery_active_entries(lport, version):
    if lport is not None and version is not None:
        return run_nsxa_ctl('ip-discovery/show-entries', [lport, version])
    elif lport is not None:
        return run_nsxa_ctl('ip-discovery/show-entries', lport)
    else:
        return run_nsxa_ctl('ip-discovery/show-entries')

def get_ip_discovery_inactive_entries(lport, version):
    if lport is not None and version is not None:
        return run_nsxa_ctl('ip-discovery/show-expired-entries',
                            [lport, version])
    elif lport is not None:
        return run_nsxa_ctl('ip-discovery/show-expired-entries', lport)
    else:
        return run_nsxa_ctl('ip-discovery/show-expired-entries')

def get_ip_discovery_stats(lport):
    if lport is not None:
        return run_nsxa_ctl('ip-discovery/show-stats', lport)
    else:
        return run_nsxa_ctl('ip-discovery/show-stats')

def get_dhcp_pending_reqs_stats(lport):
    if lport is not None:
        return run_nsxa_ctl('ip-discovery/show-dhcp-pending-request-stats',
                            lport)
    else:
        return run_nsxa_ctl('ip-discovery/show-dhcp-pending-request-stats')

def get_ip_discovery_ignore_list(lport, version):
    if lport is not None and version is not None:
        return run_nsxa_ctl('ip-discovery/show-ignore-list',
                            [lport, version])
    elif lport is not None:
        return run_nsxa_ctl('ip-discovery/show-ignore-list', lport)
    else:
        return run_nsxa_ctl('ip-discovery/show-ignore-list')

def get_ip_discovery_ignore_list_stats(lport):
    if lport is not None:
        return run_nsxa_ctl('ip-discovery/show-ignore-list-stats', lport)
    else:
        return run_nsxa_ctl('ip-discovery/show-ignore-list-stats')

# uuid VNI|VLAN vni
def get_lswitch_list(id = None):
    if id == None:
        return run_nsxa_ctl('lswitch/list')
    elif id[0]:
        return run_nsxa_ctl('lswitch/list', ['VNI', str(id[1])])
    else:
        return run_nsxa_ctl('lswitch/list', ['VLAN', str(id[1])])

def ls_uuid_to_id(uuid):
    """Converts logical switch id to uuid.

    Args:
        uuid: Logical switch UUID

    Returns:
        ID of the given logical switch in the tuple format of
        (is_vni, vni/vlan number). None if not found.
    """
    if uuid is None:
        return None
    ls_list = get_lswitch_list()
    for entry in ls_list:
        if entry[0] == uuid:
            return (entry[1] == "VNI", int(entry[2]))
    return None

def ls_id_to_uuid(id):
    """Converts logical switch id to uuid.

    Args:
        id: Logical switch id in the tuple format of
            (is_vni, vni/vlan number)

    Returns:
        UUID of the given logical switch. None if not found.
    """
    if id is None:
        return None
    ls_list = get_lswitch_list(id)
    if len(ls_list) < 1:
        return None
    return ls_list[0][0]

def id_exists(id):
    """Checks if a logical switch with given id exists.

    Args:
        id: Logical switch id in the tuple format of
            (is_vni, vni/vlan number)

    Returns:
        True if given logical switch exists, False otherwise.
    """
    return id is not None and ls_id_to_uuid(id) is not None

def get_replication_mode(vni):
    # Sample output: VNI: 12345, replication_mode: {source | mtep}
    return run_nsxa_ctl('vni/replication-mode', vni, header = False)[0][3]

def get_mac_table(vni):
    result = run_nsxa_ctl('vni/mac-vtep-label', vni)
    del result[0]
    return result

def get_vtep_list(vni):
    result = run_nsxa_ctl('vni/vtep_list', vni, split = False)
    del result[0]
    return result

def get_vtep_ip(vtep_label):
    return run_nsxa_ctl('vtep/ip', vtep_label, header= False)[0][5]

def get_vtep_mac(vtep_label):
    return run_nsxa_ctl('vtep/pmac', vtep_label, header= False)[0][6]

def get_vtep_table(vni):
    vtep_table = []
    vtep_list = get_vtep_list(vni)
    for vtep in vtep_list:
        vtep_table.append([vtep, get_vtep_ip(vtep), get_vtep_mac(vtep)])
    return vtep_table

def get_arp_table(vni):
    return run_nsxa_ctl('vni/arp-table', vni)[1:]

def get_lport_vif(lport):
    lport_vif = run_nsxa_ctl('lport/vif', lport)
    if not len(lport_vif):
       lg.error('Cant find vif for lport %s' % lport)
       return None
    return lport_vif[0][1]

def get_lport_snoop_mode(lport):
    cmd_result = run_nsxa_ctl('lport/snoop-mode', lport)[0];
    return [cmd_result[1].lower() == "true", cmd_result[2].lower() == "true",
            cmd_result[3].lower() == "true",  cmd_result[4].lower() == "true",
            cmd_result[5].lower() == "true"]

def get_lport_status(lport):
    status_table = run_nsxa_ctl('lport/status', lport)
    return status_table[0][1]

def get_lrouter_list():
    return run_nsxa_ctl("vdr/list");

def lr_id_to_uuid(id):
    try:
        return table_to_map(get_lrouter_list(), 1)[id][0][0]
    except:
        return None

def lr_uuid_to_id(uuid):
    try:
        return table_to_map(get_lrouter_list(), 0)[uuid][0][1]
    except:
        return None

def lr_uuid_to_curr_neighbors(uuid):
    try:
        return table_to_map(get_lrouter_list(), 0)[uuid][0][4]
    except:
        None

def lr_uuid_to_max_neighbors(uuid):
    try:
        return table_to_map(get_lrouter_list(), 0)[uuid][0][5]
    except:
        return None

def get_lif_list(lr_uuid = None, lif_uuid=None):
    if lr_uuid == None:
        return run_nsxa_ctl("lif/list")
    args = [lr_uuid]
    if lif_uuid:
        args.append(lif_uuid)
    return remove_header(run_nsxa_ctl("vdr/lif", args))

def get_lr_dhcp_relay_info(lr_uuid):
    args = [lr_uuid]
    lif_list = remove_header(run_nsxa_ctl("vdr/dhcp-relay", args))
    lif_dhcp_relay_map = {}
    for lif in lif_list:
        lif_dhcp_relay_map[lif[0]] = lif
    return lif_dhcp_relay_map

def get_lr_ra_config(lr_uuid):
    args = [lr_uuid]
    ra_config_list = run_nsxa_ctl("vdr/ra-config", args)
    if ra_config_list is None:
        return None
    lif_ra_config_map = {}
    for ra_config in ra_config_list:
        lif_ra_config_map[ra_config[0]] = ra_config
    return lif_ra_config_map

def get_lr_dad_config(lr_uuid):
    args = [lr_uuid]
    dad_config_list = run_nsxa_ctl("vdr/dad-config", args)
    if dad_config_list is None:
        return None
    lif_dad_config_map = {}
    for dad_config in dad_config_list:
        lif_dad_config_map[dad_config[0]] = dad_config
    return lif_dad_config_map

def get_lr_route(args):
    return remove_header(run_nsxa_ctl("vdr/route", args))

def get_lr_neighbors(lr_uuid, addr=None):
    neighbors = {}
    nbr_table = []
    for lif in get_lif_list(lr_uuid):
        args = ["neigh/show", lif[0]]
        if addr:
            args.append(addr)
        nbr_list = run_nsxa_ctl(*args)
        if nbr_list:
            nbr_table.extend(nbr_list)
            for nbr in nbr_list:
                neighbors[nbr[2]] = True
    return nbr_table;

def get_dr_vtep():
    return run_nsxa_ctl("dr-vtep/list");

def get_routing_domain_vtep_table(rd_id):
    return run_nsxa_ctl("routing-domain/vtep-table", rd_id);

def set_capture_lport(cmd_list):
    return run_nsxa_ctl("packet-capture/monitor", cmd_list)

def get_capture_sessions(cmd_list):
    return run_nsxa_ctl("packet-capture/show", cmd_list)

def del_capture_lport(cmd_list):
    return run_nsxa_ctl("packet-capture/unmonitor", cmd_list)

def del_capture_session(cmd_list):
    return run_nsxa_ctl("packet-capture/stop", cmd_list)

def get_mirror_session():
    return run_nsxa_ctl("mirror-session/list", header = False)

def get_lport_ipfix_config(cmd_list):
    return run_nsxa_ctl("lport/ipfix", cmd_list)

def get_lswitch_ipfix_config(cmd_list):
    return run_nsxa_ctl("lswitch/ipfix", cmd_list)

def get_firewall_interfaces():
    lines = run_nsxa_ctl("dfw/vif", split=False, header=False)
    res = []
    tmp = dict()
    for ln in lines:
        vals = [x.strip() for x in ln.strip().split(':')]
        # nsxa ctl server output in this format
        # "Vif ID : <vif id> \n"
        # "Port name : <port name> \n"
        # "Port number : <port num> \n"
        # "\n"
        if vals[0] == 'Vif ID':
            tmp['vifuuid'] = vals[-1]
        elif vals[0] == 'Port name':
            tmp['portname'] = vals[-1]
        elif vals[0] == 'Port number':
            tmp['portnum'] = int(vals[-1])
        elif not vals[0]:
            if not tmp:
                continue
            res.append(tmp.copy())
            tmp.clear()

    return res


def get_firewall_rules_vif(cmd_list):
    rules = run_nsxa_ctl("dfw/rules", cmd_list, split=False, header=False)

    # for more than 10k rules, the format of the list returned is:
    # ['Written to /tmp/eni-17bba439.rules.info']
    # if this is the case we just return it without processing.
    # write rules to a file
    if 'Written to' in rules[0]:
        return rules

    res = []
    for r in rules:
        vals = r.split(None, 2)
        # Sample output (first line is removed by run_nsxa_ctl):
        # "Vif ID : <vif id>\n"
        # "ruleset <ruleset uuid> {\n"
        # " rule <id> <rule description>;\n"
        # "}\n"
        # "\n"
        # "ruleset <ruleset uuid> {\n"
        # " rule <id> <rule description>;\n"
        # "}\n"
        if not vals or vals[0] == '}':
            continue
        elif vals[-1] == '{':
            res.append(dict())
            res[-1]['rules'] = []

        if vals[0] == 'ruleset':
            res[-1]['rulesetuuid'] = vals[1]
        elif vals[0] == 'rule':
            # the rule description ends with a semicolon ";"
            res[-1]['rules'].append({'ruleid': int(vals[1]),
                                    'rule': vals[2][:-1]})

    return res


def get_firewall_address_set_name(cmd_list):
    return remove_footer(run_nsxa_ctl("dfw/addrset", cmd_list, split=False))


def get_firewall_address_sets_vif(cmd_list):
    return run_nsxa_ctl("dfw/addrsets", cmd_list, split=False, header=False)


def get_firewall_exclude_vifs():
    return remove_footer(run_nsxa_ctl("dfw/exclude", split=False))


def get_gateway_connection_status():
    return run_nsxa_ctl("gateway/show", split=False, header=False)

def get_gateway_certificates(ip):
    return run_nsxa_ctl("gateway-cert/show", ip, split=False, header=False)

def del_gateway_certificates(ip):
    return run_nsxa_ctl("gateway-cert/del", ip, split=False, header=False)


def get_gateway_fullsync_status():
    return run_nsxa_ctl("full-sync/status", split=False, header=False)


def get_vif_list():
    vif_lport_list = run_nsxa_ctl('lport/vif', [])
    vifs = [ [vif_lport[1]] for vif_lport in vif_lport_list ]

    return vifs

def get_vif_id():
    vif_ids = []
    ovs_vsctl_cmd = os.path.join(FilePaths.OVS_BIN_DIR, 'ovs-vsctl')
    ports = run_command([ovs_vsctl_cmd, "list-ports", "nsx-managed"],
                        False, False)
    if not ports:
        lg.error("unable to retrieve the ports")
        return vif_ids
    for port in ports:
        res = run_command([ovs_vsctl_cmd, "--if-exists", "get",
                          "interface", port, "external_ids:iface-id"],
                          False, False)
        if not res:
            continue
        vif_ids.append(res[0].strip('"'))

    return vif_ids


def get_lport_on_vif(vif):
    vif_lport_list = run_nsxa_ctl('lport/vif', [])
    for vif_lport in vif_lport_list:
       if vif_lport[1] == vif:
          return True, vif_lport[0]
    return False, None


def get_logical_switches_on_vm():
    ls_vm = []
    ls_list = get_lswitch_list()  # uuid vni
    vifs = get_vif_id()  # returns a list
    if not vifs:
        lg.error("unable to retrieve vif ID")
        return ls_vm
    for ls in ls_list:
        for p in get_lport_list(ls_uuid_to_id(ls[0])):
            if get_lport_vif(p[0]) in vifs:
                ls_vm.append(ls)
                break
    return ls_vm


def get_logical_routers_on_vm():
    lr_list = get_lrouter_list()
    levels = 3  # get 3 levels of routers
    lr_vm = set()
    vni_lr_map = dict()
    lr_vni_map = dict()
    for lr in lr_list:
        lr_uuid = lr[0]
        lr_vni_map[lr_uuid] = set()
        for p in get_lif_list(lr_uuid):
            vni = int(p[3])
            lr_vni_map[lr_uuid].add(vni)
            if vni not in vni_lr_map:
                vni_lr_map[vni] = set()
            vni_lr_map[vni].add(lr_uuid)
    vnis = {int(ls[2])
            for ls in get_logical_switches_on_vm() if ls[1] == "VNI"}
    vni_visited = set()
    for i in range(levels):
        if not vnis:
            break
        lr_next = set()
        for v in vnis:
            lrs = vni_lr_map.get(v)
            if not lrs:
                continue
            for lr in lrs:
                if lr not in lr_vm:
                    lr_next.add(lr)
            vni_visited.add(v)
        lr_vm.update(lr_next)
        vnis.clear()
        for l in lr_next:
            for v in lr_vni_map.get(l):
                if v not in vni_visited:
                    vnis.add(v)

    return [lr for lr in lr_list if lr[0] in lr_vm]


def get_firewall_context_profiles():
    return remove_footer(run_nsxa_ctl("dfw/profiles", split=False, header=False))


def get_firewall_context_profile_fqdn(cmd_list):
    return remove_footer(run_nsxa_ctl("dfw/profile-fqdn", cmd_list, split=False, header=False))
