# ***************************************************************************
# Copyright 2014-2020 VMware, Inc.  All rights reserved. VMware Confidential.
# ***************************************************************************
# $Id$
# $DateTime$
# $Change$
# $Author$
# ***************************************************************************


__all__ = ["CliArgumentDescriptor"]


import nsx_logging
import getpass
import os
import re
try:
    import simplejson
except ImportError:
    import json as simplejson
import yaml

from lib.json_schema import JsonSchema
from cli.utils.constants import CENTRAL_CLI_END_TOKEN
from cli.utils.constants import CENTRAL_CLI_STR
from cli.descriptors.descriptor import Descriptor
from cli.descriptors.descriptor import SCHEMA_REGISTRY
from cli.utils.central_cli_api import CentralCLIAPIException
from cli.utils.central_cli_api import get_arg_values
from cli.utils.central_cli_api import verify_arg
from cli.utils.central_cli_utils import is_central_cli_command
from cli.utils.central_cli_utils import get_central_cli_nodes
from cli.utils.central_cli_utils import get_central_cli_user_input
from cli.utils.import_utils import get_class
from cli.utils.render_utils import render_definition
from cli.utils.rest_utils import local_node_request
from cli.utils.error_code import ErrorCode as ec
from compat23.compat23_utils import compat23_basestring


lg = nsx_logging.getLogger("cli.descriptors.cli_argument_descriptor")


CLI_BASE_DIR = os.path.dirname(os.path.dirname(__file__))
CLI_ARGUMENT_SCHEMA_FILE = os.path.join(os.path.dirname(__file__),
                                       'cli_argument_descriptor.yml')
CLI_ARGUMENT_TYPES = dict((r.id, r)
                         for r in yaml.load(open(CLI_ARGUMENT_SCHEMA_FILE), Loader=yaml.Loader))
for key in list(CLI_ARGUMENT_TYPES.keys()):
    if key not in SCHEMA_REGISTRY:
        JsonSchema(CLI_ARGUMENT_TYPES[key].raw_schema,
                   schema_registry=SCHEMA_REGISTRY, type_path=[key],
                   strict=True)


def is_argument(token):
    """Determing if specified string token is a CLI argument.

    Args:
        token: String token to test if argument.
    Return:
        True if speciifed token is an argument; otherwise, False.
    """
    return token.startswith("<") and token.endswith(">")


def resolve_argument_name(name):
    """Return name of argument.

    Args:
        name: Argument name with or without <> prefix/suffix.
    Return:
        Argument name without the <> prefix/suffix.
    """
    if not is_argument(name):
        return name
    return name[1:-1]


def show_argument_description(arg_desc):
    """Show argument description as defined in its descriptor yaml.

    Args:
        arg_desc: Argument descriptor.
    """
    desc = arg_desc.description
    if desc:
        render_definition(arg_desc.label, desc)


class CliArgumentDescriptor(Descriptor):
    """Descriptor specification of a CLI argument."""
    yaml_tag = '!CliArgument'

    def __init__(self, id, schema_spec):
        super(CliArgumentDescriptor, self).__init__(id, schema_spec)
        self._factory = None  # Set by factory
        self._raw_schema = schema_spec
        self._name = schema_spec["name"]
        self._type = schema_spec["type"]
        self._label = schema_spec["label"]
        self._pattern = schema_spec.get("pattern")
        self._full_match_required = schema_spec.get("full_match_required", False)
        self._enum = schema_spec.get("enum")
        self._module = schema_spec.get("module")
        self._class_name = schema_spec.get("class_name")
        self._description_module = schema_spec.get("value_description_module")
        self._description_class_name = schema_spec.get(
            "value_description_class_name")
        self._nsx_feature = schema_spec.get("nsx_feature")
        self._limit = self._raw_schema.get("limit", 1)
        self._ending_word = self._raw_schema.get("ending_word")
        self._obfuscate_in_audit_log = self._raw_schema.get(
                                            "obfuscate_in_audit_log", False)

        self._is_subset_of = schema_spec.get("is_subset_of", [])

        # This property should not be mixed with 'verify_arg' enum value
        self._verify = schema_spec.get("verify", True)

        if (self._limit > 1) and (not self._ending_word):
            raise DescriptorException("ending_word is required for limit > 1")

        if self._type == "enum_arg" and self._enum is None:
            msg = ("Enum argument '%s' did not specify an enum property" %
                   self._name)
            lg.error(msg, error_code=ec.INVALID_ARGUMENT)
            assert False, msg
        elif self._type == "pattern_arg" and self._pattern is None:
            msg = ("Pattern argument '%s' did not specify a pattern "
                   "property" % self._name)
            lg.error(msg, error_code=ec.INVALID_ARGUMENT)
            assert False, msg
        elif (self._type == "query_arg" and
              (self._module is None or self._class_name is None)):
            msg = ("Query argument '%s' did not specify a module or class_name"
                   % self._name)
            lg.error(msg, error_code=ec.INVALID_ARGUMENT)
            assert False, msg
        elif (self._type == "verify_arg" and
              (self._module is None or self._class_name is None)):
            msg = ("Verify argument '%s' did not specify a module or class_name"
                   % self._name)
            lg.error(msg, error_code=ec.INVALID_ARGUMENT)
            assert False, msg
        elif (self._type not in
              ["enum_arg", "pattern_arg", "query_arg", "verify_arg"]):
            msg = ("Unexpected argument type '%s' for arg name '%s'" %
                   (self._type, self._name))
            lg.error(msg, error_code=ec.INVALID_ARGUMENT)
            assert False, msg

    @property
    def name(self):
        return self._name

    @property
    def label(self):
        return self._label

    @property
    def help_summary(self):
        return self._raw_schema.get("help_summary", "")

    @property
    def description(self):
        return self._raw_schema.get("description", "")

    @property
    def full_match_required(self):
        return self._full_match_required

    @property
    def enum_argument(self):
        return self._type == "enum_arg"

    @property
    def pattern_argument(self):
        return self._type == "pattern_arg"

    @property
    def query_argument(self):
        return self._type == "query_arg"

    @property
    def verify_argument(self):
        return self._type == "verify_arg"

    @property
    def limit(self):
        return self._limit

    @property
    def ending_word(self):
        return self._ending_word

    @property
    def is_multiarg(self):
        """Whether argument allows multiple tokens as allowed value"""
        # TODO This can be more involved, specially with ending_word and defaults
        # It's a syntax error if limit > 1 and ending_word not specified
        return True if self.ending_word else False

    @property
    def obfuscate_in_audit_log(self):
        return self._obfuscate_in_audit_log

    @property
    def is_subset_of(self):
        """All values of this argument type are subset of given list of other argument types"""
        return self._is_subset_of

    @property
    def verify(self):
        """Control flag for argument verification in auto-complete"""
        return self._verify

    def get_allowed_values(self, req_dict,
                           empty_list_if_error=True, debug_mode=False,
                           session_nodes=[]):
        """Return allowed values by argument, or [] if it can't be predetermined

        Args:
            req_dict:  Dict with request details like cmd_list, context.
            empty_list_if_error: If a query argument returns an error, return an
                empty list instead.
            session_nodes: If running in Central CLI's session mode, this list
                contains all remote nodes for which possible allowed values are
                being requested.
        Return:
            List of allowed values if self is either a enum or query argument;
            otherwise, an empty list.
        """

        result = []
        if self.enum_argument and self._enum:
            result = self._enum
        elif self.pattern_argument or self.verify_argument:
            result = []
        elif self.query_argument and self._module and self._class_name:
            # If it's Central CLI command
            # Unit tests may not supply cmd_list in req_dict
            if 'cmd_list' in req_dict and len(req_dict['cmd_list']) > 0:
                # Only get arg values from remote, if it's Central CLI command,
                # and argument is not the one which gathers all remote nodes
                if ((is_central_cli_command(req_dict['cmd_list'])
                        or session_nodes)
                    and self.ending_word != CENTRAL_CLI_END_TOKEN):
                    return self.get_remote_allowed_values(req_dict,
                                                          empty_list_if_error,
                                                          debug_mode,
                                                          session_nodes)
            # Try to get value from factory verify arg cache
            result = self._factory.verify_arg_cache.get_arg_values(req_dict,
                                                                   self.id)
            if result:
                return result

            try:
                cls = get_class(self._module, self._class_name)
            except ImportError:
                msg = ("Error resolving query argument '%s' module '%s' or "
                       "class_name '%s'" %
                       (self._name, self._module, self._class_name))
                lg.exception(msg, error_code=ec.INVALID_ARGUMENT)
                assert False, msg
                return []
            json_query_args = None
            try:
                json_query_args = cls(self)(req_dict)
                result = simplejson.loads(json_query_args)
                if empty_list_if_error and isinstance(result, str):
                    result = []
            except ValueError:
                msg = ("json conversion error converting query argument '%s' "
                       "value: %s" % (self._name, json_query_args))
                lg.exception(msg, error_code=ec.INVALID_ARGUMENT)
                assert False, msg

            self._factory.verify_arg_cache.set_arg_values(req_dict, self.id,
                                                          result)

        if self.is_multiarg: # Append ending_word, if argument is multi_args
            result.append(self.ending_word)

        return result

    def get_remote_allowed_values(self, req_dict, empty_list_if_error=True,
                                  debug_mode=False, session_nodes=[]):
        """Return allowed values by argument, or [] on remote nodes

        information about remote nodes is extracted from supplied req_dict.

        Args:
            req_dict:  Dict with request details like cmd_list, context.
            empty_list_if_error: If a query argument returns an error, return an
                empty list instead.
            session_nodes: If running in Central CLI's session mode, this list
                contains all remote nodes for which possible allowed values are
                being requested.
        Return:
            List of allowed values if self is either an enum or query argument;
            otherwise, an empty list.
        """
        context = req_dict['context']
        if session_nodes:
            nodes = session_nodes
            user_input = req_dict['cmd_list']
        else:
            nodes = get_central_cli_nodes(req_dict['cmd_list'])
            user_input = get_central_cli_user_input(req_dict['cmd_list'])

        user_input = ' '.join(user_input).strip()

        result = set()

        for node in nodes:
            try:
                node_uuid = self._factory.node_cache[node]['uuid']
                node_type = self._factory.node_cache[node]['node_type']
                output = get_arg_values(node_uuid, node_type, context, self.id,
                                        user_input, "", getpass.getuser(),
                                        debug_mode)
                for o in output:
                    result.add(o['value'])
            except CentralCLIAPIException:
                err = "Failed to get arg allowed values for node: %s" % node
                if not empty_list_if_error:
                    lg.error(err, error_code=ec.OPERATION_FAILED)
                    assert False, err
                else:
                    lg.warn(err)

        return list(result)

    def match_verify_arg_value(self, value, req_dict, false_if_error=True,
                               debug_mode=False, remote_nodes=None):
        """Return True if self is a verify argument and value is valid.

        Args:
            value:  Value to verify if self is a verify argument.
            req_dict:  Dict with request details like cmd_list, context.
            false_if_error: If verify argument returns an error, return False
                instead.
            debug_mode: is debug set for current cli session
            remote_nodes: Incase of active session mode, list of nodes with
                node-uuid(s) or None.
        Return:
            True if self is a verify argument and specified value is valid;
            otherwise, False.
        """
        result = False
        if not self.verify_argument:
            return False

        # If it's Central CLI command or an explicit specification of remote nodes
        if (is_central_cli_command(req_dict['cmd_list']) or remote_nodes)\
           and self.id != 'RegisteredNodeUuid':
            return self.match_remote_verify_arg_value(value,
                                                      req_dict,
                                                      false_if_error,
                                                      debug_mode,
                                                      remote_nodes)

        try:
            cls = get_class(self._module, self._class_name)
        except ImportError:
            msg = ("Error resolving verify argument '%s' module '%s' or "
                   "class_name '%s'" %
                   (self._name, self._module, self._class_name))
            lg.exception(msg, error_code=ec.INVALID_ARGUMENT)
            assert False, msg
            return False
        try:
            result = cls().verify(value, req_dict)
            if false_if_error and isinstance(result, str):
                result = False
        except Exception:
            msg = ("Error matching verify argument value %s: %s" %
                   (self._name, value))
            lg.exception(msg, error_code=ec.INTERNAL_ERROR)
            assert False, msg
            return False
        return result

    def match_remote_verify_arg_value(self, value, req_dict,
                                      false_if_error=True, debug_mode=False,
                                      remote_nodes=None):
        """Return True if self is a verify argument and value is valid on any of
        the remote node

        Args:
            value:  Value to verify if self is a verify argument.
            req_dict:  Dict with request details like cmd_list, context.
            false_if_error: If verify argument returns an error, return False
                instead.
            debug_mode: is debug set for current cli session
            remote_nodes: Incase of active session mode, list of nodes with
                node-uuid(s) or None.
        Return:
            True if self is a verify argument and specified value is valid;
            otherwise, False.
        """
        if remote_nodes is None:
            remote_nodes = []

        context = req_dict['context']
        nodes = get_central_cli_nodes(req_dict['cmd_list']) or remote_nodes
        user_input = get_central_cli_user_input(req_dict['cmd_list'])
        user_input = ' '.join(user_input).strip()

        result = []
        for node in nodes:
            try:
                node_uuid = self._factory.node_cache[node]['uuid']
                node_type = self._factory.node_cache[node]['node_type']
                output = verify_arg(node_uuid, node_type, context,
                                    self.id.replace(CENTRAL_CLI_STR, ''),
                                    user_input, value, getpass.getuser(),
                                    debug_mode)
                result.append(output)
            except (CentralCLIAPIException, KeyError):
                err = "Failed to verify arg values for node: %s" % node
                if not false_if_error:
                    lg.error(err, error_code=ec.OPERATION_FAILED)
                    assert False, err
                else:
                    lg.warn(err)
                    return False
        return any(result) # Return True, if this succeed for any node

    def get_argument_value_desc(self, value, req_dict,
                                strip_trailing_punc=True):
        """Get argument value description.

        Args:
            value:  Argument value. If specified, invoke descriptor's value
                description class (if defined) to get value description.
            req_dict:  Dict with request details like cmd_list, context.
            strip_trailing_punc:  True, if value description's trailing period
                should be stripped.
        """
        desc = self.description  # Default to description description property
        if (value is not None and self._description_module is not None and
                self._description_class_name is not None):
            try:
                cls = get_class(
                    self._description_module, self._description_class_name)
                desc = cls().value_description(value, req_dict, self._factory)
            except ImportError:
                msg = ("Error importing argument '%s' value description, "
                       "value module '%s' or value class_name '%s'" %
                       (self.name, self._module, self._class_name))
                lg.exception(msg, error_code=ec.INVALID_ARGUMENT)
            except Exception:
                msg = ("Error resolving argument '%s' value description, "
                       "value module '%s' or value class_name '%s'" %
                       (self.name, self._module, self._class_name))
                lg.exception("Error getting argument value description",
                             error_code=ec.INTERNAL_ERROR)
        if desc and strip_trailing_punc:
            desc = desc.strip()
            if len(desc) > 0 and desc[-1] in ["."]:
                desc = desc[:-1]
        return desc

    def sort_key(self, item):
        """Key param, for overriding default sorting logic.

        This function can be used as a comparison function, which is equivalent
        to key parameter to Python's sort function. Currently this is used for
        command <tab><tab> auto-completion, to allow arguments to provide a
        unique way to sort completions. Central CLI is using this to group
        various nodes based on their type (i.e. esx or mgr or edge).
        """
        if not isinstance(item, (compat23_basestring, bytes)):
            try:
                return tuple(['' if i is None else i for i in item])
            except TypeError:
                return item
        return item
