import nsx_logging
import os
import re
import socket
import sys

sys.path.append('/opt/vmware/nsx-node-api/bin/python')
try:
    from management_api.napi.error_code import ErrorCode as ec
except ImportError:
    pass

lg = nsx_logging.getLogger(__name__)

_RSYSLOG_CONF = "/etc/rsyslog.conf"

_EXPORTER_SIG = "# nsx exporter: "
_IF_BEGIN_SIG = "if "
_IF_END_SIG = "}  %s" % _EXPORTER_SIG

_IF_AND_TOKEN = " and "
_IF_OR_TOKEN = " or "
_MSGID_TOKEN = "$msgid == "
_SDATA_TOKEN = "$structured-data contains "

_MSGID_NAME_RE_TEXT = ".*\\$msgid[\\s]+\\=\\=[\\s]+\\'(.+)'.*"
_MSGID_NAME_RE = re.compile(_MSGID_NAME_RE_TEXT)
_SDATA_NAME_RE_TEXT = ".*\\$structured-data[\\s]+contains[\\s]+\\'(.+)'.*"
_SDATA_NAME_RE = re.compile(_SDATA_NAME_RE_TEXT)
_EXPORTER_RE_TEXT = ("^[\\s]*(.+)\\.(emerg|alert|crit|err|warning|notice|info|"
                     "debug)[\\s]+([\\@]{1,2})(.+)\\:([\\d]+);.+%s(.+)$" %
                     _EXPORTER_SIG)
_EXPORTER_RE = re.compile(_EXPORTER_RE_TEXT)

_SIG_TLS_CA_FILE = "$DefaultNetstreamDriverCAFile"
_SIG_TLS_CERT_FILE = "$DefaultNetstreamDriverCertFile"
_SIG_TLS_KEY_FILE = "$DefaultNetstreamDriverKeyFile"
_TLS_CA_DIR = "/config/vmware/nsx-node-api/syslog"

_EXPORTER_FIREWALL_RULES_FILE = os.path.join(_TLS_CA_DIR,
                                             "exporter_firewall_rules")


def get_rsyslog_conf():
    """Return path to rsyslog.conf"""
    return _RSYSLOG_CONF


def read_rsyslog_conf():
    """Return lines in /etc/rsyslog.conf or None if an error occurred."""
    try:
        with open(get_rsyslog_conf()) as fd:
            return fd.read().split("\n")
    except Exception:
        try:
            lg.exception("Error reading %s" % get_rsyslog_conf(),
                     error_code=ec.FAIL_READ_FILE)
        except:
            lg.warn("Error reading %s" % get_rsyslog_conf())
    return None


def read_pem_file(ca_file):
    try:
        with open(ca_file) as fd:
            return fd.read()
    except Exception:
        raise


def extract_exporter_properties(exporter_config):
    """Return exporter property dictionary given exporter config line.

    Note: the calling method needs to do additional processing to include
          TLS, facility and messageid properties.
    """
    match = re.match(_EXPORTER_RE, exporter_config)
    if match is None:
        return None
    facilities = match.group(1)
    level = match.group(2)
    protocol = match.group(3)
    server = match.group(4)
    port = match.group(5)
    exporter_name = match.group(6)
    exporter_prop = {
        "exporter_name": exporter_name,
        "level": level.upper(),
        "port": int(port),
        "protocol": "TCP" if protocol == "@@" else "UDP",
        "server": server}
    if facilities != "*":
        exporter_prop["facilities"] = facilities.upper().split(",")
        try:
            from management_api.napi.rest_routine_utils import set_meta
            set_meta(exporter_prop,
                "/node/services/syslog/exporters/%s" % exporter_name,
                "NodeSyslogExporterProperties")
        except ImportError:
            pass
    return exporter_prop


def extract_exporter_msgids(msgid_config):
    """Return list of MSGIDs given exporter msgid config line."""
    if _MSGID_TOKEN not in msgid_config:
        return None
    msgids = []
    if _IF_AND_TOKEN in msgid_config:
        msgid_config = msgid_config.split(_IF_AND_TOKEN)[0]
    parts = msgid_config.split(_IF_OR_TOKEN)
    for part in parts:
        match = re.match(_MSGID_NAME_RE, part)
        if match:
            msgids.append(match.group(1))
    return msgids


def extract_exporter_sdata(sdata_config):
    """Return list of structured data given exporter sdata config line."""
    if _SDATA_TOKEN not in sdata_config:
        return None
    sdata = []
    if _IF_AND_TOKEN in sdata_config:
        sdata_config = sdata_config.split(_IF_AND_TOKEN)[1]
    parts = sdata_config.split(_IF_OR_TOKEN)
    for part in parts:
        match = re.match(_SDATA_NAME_RE, part)
        if match:
            sdata.append(match.group(1))
    return sdata


def read_exporter_properties(exporter_name=None, rsyslog_conf_lines=None):
    """Return exporter properties optionally filtered by exporter_name.

    Args:
        exporter_name: Name of specified exporter to read, if None, return all.
        rsyslog_conf_lines: lines in /etc/rsyslog.conf, if None, this method
            will read /etc/rsyslog.conf.
    Returns:
        List of configured syslog exporters optionally filtered by
        exporter_name, or None if an error occurred.
    """
    # /etc/rsyslog.conf configures exporter either via a single line (i.e.
    # no MSGID specified for exporter) or via multiple lines (i.e. one or
    # more MSGIDs specified for exporter). For example:
    #
    # *.err @@127.0.0.1:514;RFC5424fmt  # nsx exporter: exporter1
    #
    # if $msgid == 'aaaa' or $msgid == 'bbbb' then {  # nsx exporter: exporter2
    # local6,local7.crit @127.0.0.1:514;RFC5424fmt  # nsx exporter: exporter2
    # }  # nsx exporter: exporter2
    #
    # Note that NSX adds "# nsx exporter: <exporter_name>" to each exporter
    # it configures.
    result = []
    lines = rsyslog_conf_lines or read_rsyslog_conf()
    if lines is None:
        return None
    sig = ("%s%s" % (_EXPORTER_SIG, exporter_name)
           if exporter_name else _EXPORTER_SIG)
    multi_line_exporter = []
    tls_ca_pem = None
    tls_client_ca_pem = None
    tls_cert_pem = None
    tls_key_pem = None
    for line in lines:
        if sig in line:
            # Current line in /etc/rsyslog contains nsx exporter signature,
            # we can be in one of the following possible scenarios:
            # 1. start of a multi-line exporter
            # 2. at end of a multi-line exporter
            # 3. in the middle of a multi-line exporter
            # 4. current line is a single line exporter
            # 5. current line is a TLS configuration line
            if sig in line and line.startswith(_IF_BEGIN_SIG):  # scenario 1
                multi_line_exporter = [line]
            elif multi_line_exporter and line.startswith(_IF_END_SIG):  # 2
                if len(multi_line_exporter) < 2:
                    try:
                        lg.error("Invalid rsyslog.conf configuration:\n%s",
                             "\n".join(multi_line_exporter),
                             error_code=ec.INVALID_CONFIG)
                    except:
                        lg.warn("Invalid rsyslog.conf configuration:\n%s",
                             "\n".join(multi_line_exporter))
                    return None
                exporter_prop = extract_exporter_properties(
                    multi_line_exporter[1])
                if not exporter_prop:
                    try:
                        lg.error("Invalid syslog exporter configuration: %s",
                             multi_line_exporter[1],
                             error_code=ec.INVALID_CONFIG)
                    except:
                        lg.warn("Invalid syslog exporter configuration: %s",
                             multi_line_exporter[1])
                    multi_line_exporter = []
                    continue  # ignore malformed exporter
                msgids = extract_exporter_msgids(multi_line_exporter[0])
                sdata = extract_exporter_sdata(multi_line_exporter[0])
                if msgids:
                    exporter_prop["msgids"] = msgids
                if sdata:
                    exporter_prop["structured_data"] = sdata
                if tls_ca_pem:
                    exporter_prop["protocol"] = "TLS"
                    exporter_prop["tls_ca_pem"] = tls_ca_pem
                if tls_cert_pem:
                    exporter_prop["tls_cert_pem"] = tls_cert_pem
                if tls_key_pem:
                    exporter_prop["tls_key_pem"] = tls_key_pem
                result.append(exporter_prop)
                tls_ca_pem = None
                tls_client_ca_pem = None
                tls_cert_pem = None
                tls_key_pem = None
                multi_line_exporter = []
            elif multi_line_exporter:  # scenario 3
                multi_line_exporter.append(line)
            elif line.strip()[0] == "$":
                parts = line.split()
                if parts[0] == _SIG_TLS_CA_FILE:
                    ca_file = parts[1]
                    try:
                        tls_ca_pem = read_pem_file(ca_file)
                    except Exception:
                        try:
                            lg.exception("Error opening exporter CA pem file: %s",
                                     ca_file, error_code=ec.FAIL_OPEN_FILE)
                        except:
                            lg.warn("Error opening exporter CA pem file: %s",
                                    ca_file)
                elif parts[0] == _SIG_TLS_CERT_FILE:
                    cert_file = parts[1]
                    try:
                        tls_cert_pem = read_pem_file(cert_file)
                    except Exception:
                        try:
                            lg.exception("Error opening exporter cert pem file: %s",
                                     cert_file, error_code=ec.FAIL_OPEN_FILE)
                        except:
                            lg.warn("Error opening exporter cert pem file: %s",
                                    cert_file)
                elif parts[0] == _SIG_TLS_KEY_FILE:
                    key_file = parts[1]
                    try:
                        tls_key_pem = read_pem_file(key_file)
                    except Exception:
                        try:
                            lg.exception("Error opening exporter key pem file: %s",
                                     key_file, error_code=ec.FAIL_OPEN_FILE)
                        except:
                            lg.warn("Error opening exporter key pem file: %s",
                                    key_file)
            else:  # scenario 4
                exporter_prop = extract_exporter_properties(line)
                if not exporter_prop:
                    try:
                        lg.error("Invalid syslog exporter configuration: %s", line,
                             error_code=ec.INVALID_CONFIG)
                    except:
                        lg.warn("Invalid syslog exporter configuration: %s", line)
                    return None
                if tls_ca_pem:
                    exporter_prop["protocol"] = "TLS"
                    exporter_prop["tls_ca_pem"] = tls_ca_pem
                if tls_cert_pem:
                    exporter_prop["tls_cert_pem"] = tls_cert_pem
                    # If exporter name is not given, need to extract it to
                    # access the client CA file
                    if not exporter_name:
                        exporter_name = line.split(sig)[1]
                    client_ca_path = os.path.join(_TLS_CA_DIR,
                        "%s_client_ca.pem" % exporter_name)
                    try:
                        tls_client_ca_pem = read_pem_file(client_ca_path)
                    except Exception:
                        try:
                            lg.exception("Error opening exporter client CA pem "
                                     "file: %s", client_ca_path,
                                     error_code=ec.FAIL_OPEN_FILE)
                        except:
                            lg.warn("Error opening exporter client CA pem "
                                     "file: %s", client_ca_path)
                    exporter_prop["tls_client_ca_pem"] = tls_client_ca_pem
                if tls_key_pem:
                    exporter_prop["tls_key_pem"] = tls_key_pem
                result.append(exporter_prop)
                tls_ca_pem = None
                tls_client_ca_pem = None
                tls_cert_pem = None
                tls_key_pem = None
    return result


def get_all_existing_firewall_rules():
    # Read exporter_firewall_rules file for existing iptable rules
    results = {}
    try:
        if os.path.isfile(_EXPORTER_FIREWALL_RULES_FILE):
            with open(_EXPORTER_FIREWALL_RULES_FILE, 'r') as f:
                for line in f:
                    tokens = line.split()
                    # Dictionary format: (<ip>, <port>, <protocol>): <hostname>
                    results[(tokens[1], tokens[2], tokens[3])] = tokens[0]
    except Exception:
        try:
            lg.exception("Error reading %s" % _EXPORTER_FIREWALL_RULES_FILE,
                      error_code=ec.FAIL_READ_FILE)
        except:
            lg.warn("Error reading %s" % _EXPORTER_FIREWALL_RULES_FILE)
    return results


def get_firewall_rules_for_exporters(exporters):
    # Read current configs and collect corresponding iptable rules
    results = {}
    for exporter in exporters:
        ip_addr = socket.gethostbyname(exporter['server'])
        try:
            hostname = socket.gethostbyaddr(exporter['server'])[0]
        except socket.herror:
            hostname = exporter['server']
        # Dictionary format: (<ip>, <port>, <protocol>): <hostname>
        results[(ip_addr, str(exporter['port']),
            "TCP" if exporter['protocol'] is "TLS"
                else exporter['protocol'])] = hostname
    return results
