# Copyright (C) 2016, 2017 VMware, Inc. All rights reserved.

# TODO Removing the next line (for python3 compatibiltiy)
# causes unit tests to fail as the keepalive routine relies
# on the division operator.
from __future__ import division
from contextlib import contextmanager
import collections
import gevent
import gevent.lock
from gevent import socket, ssl
import logging
import uuid

from vmware.nsx.rpc.client.bin_stream import BinStream
from vmware.nsx.rpc.client import channel
from vmware.nsx.rpc import nsx_rpc_pb2
from vmware.nsx.rpc.status import CallStatus
from vmware.nsx.rpc.status import CloseReason

LOGGER = logging.getLogger("nsxrpc")

# default time in seconds for the keepalive timeout (client and server)
KEEPALIVE_CLIENT_DEFAULT_TIMEOUT_SEC = 60
KEEPALIVE_SERVER_DEFAULT_TIMEOUT_SEC = 120
# number of pings that will be send in KEEPALIVE_TIMEOUT_SEC
KEEPALIVE_NUM_PINGS = 3


class NsxRpcMsgHeader(object):
    '''
      | 1 byte of flags | 1 byte length of serialized StreamControlMsg |
      | 2 byte length of serialized FrameMsg |
      | 4 byte length of serialized payload |
      | 16 bytes stream_id |
      | serialized StreamControlMsg |
      | serialized FrameMsg |
      | serialized payload |

      The lengths should be carried in little endian byte order.
    '''
    SIZE = 24

    def __init__(self, flags, stream_control_size,
                 frame_size, payload_size, stream_id):
        self.flags = flags
        self.stream_control_size = stream_control_size
        self.frame_size = frame_size
        self.payload_size = payload_size
        self.stream_id = stream_id

    @staticmethod
    def decode(stream):
        assert isinstance(stream, BinStream)
        assert len(stream) >= NsxRpcMsgHeader.SIZE

        flags = stream.read_u8()
        stream_control_size = stream.read_u8()
        frame_size = stream.read_u16()
        payload_size = stream.read_u32()
        stream_id = uuid.UUID(bytes=stream.read_data(16))
        return NsxRpcMsgHeader(flags, stream_control_size,
                               frame_size, payload_size, stream_id)

    def encode(self, stream):
        assert isinstance(stream, BinStream)

        assert(self.stream_control_size < 0x100)
        assert(self.frame_size < 0x10000)

        stream.write_u8(self.flags)
        stream.write_u8(self.stream_control_size)
        stream.write_u16(self.frame_size)
        stream.write_u32(self.payload_size)
        stream.write_data(self.stream_id.bytes)

    def __str__(self):
        return ('NsxRpcMsgHeader {flags: %u, stream_control_size: %u, '
                'frame_size: %u, payload_size: %u, '
                'stream_id: %s}') % (
            self.flags, self.stream_control_size,
            self.frame_size, self.payload_size, str(self.stream_id))


@contextmanager
def accept_one(endpoint):
    """ DEPRECATED API, only for use by NSX Container Plugin.

    Creates a connection and calls AcceptOne(endpoint) on it.
    Exceptions thrown from within the if block are propagated as is.

    Can be used as:
        with accept_one('tcp://localhost:2688') as connection:
            # use the accepted connection.
    """
    connection = NsxRpcConnection()
    connection.AcceptOne(endpoint)
    try:
        yield connection
    finally:
        connection.Close(CloseReason(CloseReason.LOCAL_CLOSE))


RetryPolicy = collections.namedtuple('RetryPolicy', ['attempts', 'delay'])


class NsxRpcConnection(object):
    """NsxRpcConnection represents an NSX RPC connection. It holds a socket
    and can send and receive NSX RPC messages.

    User must explicitly call Close() once the connection is not needed
    anymore.
    """

    def __init__(self):
        # sock being None signifies that connection is closed
        self.sock = None
        # lock to synchronize access to sock.send*().
        self.send_lock = gevent.lock.Semaphore(1)

        # save parameters for reconnection attempts
        self.endpoint = None
        self.sec_ctx = None

        # We don't have a use case for endpoint ids since we don't
        # support rpc servers in python yet. There are also no tests.
        # Don't expose them.
        self.endpoint_ids = []

        # {stream_id -> NsxRpcChannel}
        self.channels = {}

        # Once we connect, receiver will have the greenlet receiving
        # from the socket.
        self.receiver = None

        # Once we detect that the connection went down, spawn a greenlet
        # that attempts to reconnect.
        self.reconnect_routine = None

        # number of reconnect attempts and delay between them
        self.retry_policy = None

        # If we started the connection, keep_alive_routine will have the
        # greenlet sending keep alive messages
        self.keep_alive_routine = None

        # expected keepalive timeout, in seconds; this is initialized
        # to its proper value just before we spawn the keepalive
        # greenlet, depending on whether we are client or server
        self.keep_alive_timeout_sec = 0

        self.remaining_keep_alive_budget = self.keep_alive_timeout_sec

        # number of keep alive pings to send in keep_alive_timeout_sec
        self.keep_alive_num_pings = KEEPALIVE_NUM_PINGS

        # keepalive sequence id
        self.keep_alive_counter = 0

        # keep track of last received keepalive id
        self.keep_alive_last_recv_id = 0

    @property
    def is_active(self):
        """Return true if the connection is active - has been
        connected/accepted and not yet closed."""
        return self.sock is not None

    def RegisterChannel(self, stream_id, channel):
        """Register channel with stream_id as being a user of this connection.
        """

        LOGGER.debug("NsxRpcConnection (%s) registering channel for "
                     "stream_id %s", self, stream_id)
        self.channels[stream_id] = channel

    def Connect(self, endpoint, sec_ctx=None, retry_policy=None):
        """ Attempts to connect to endpoint with the given parameters.

        Previous connections should be disposed by calling Close() before
        calling this function.

        If a retry policy is specified, the connection will attempt to recover
        when a network error occurs by trying to reconnect to the endpoint.
        """
        LOGGER.debug("NsxRpcConnection (%s) Connect(%s)", self, endpoint)
        self.endpoint = endpoint
        self.sec_ctx = sec_ctx
        self.retry_policy = retry_policy

        if retry_policy is not None:
            # there are no other tasks running at this point so we simply loop
            # on the main thread until a connection is established
            self._ConnectRetry(endpoint, sec_ctx)
        else:
            self._Connect(endpoint, sec_ctx)

        # spawn background tasks (receiver and keep alive)
        self._OnConnect()

    def _Connect(self, endpoint, sec_ctx=None):
        if endpoint.startswith("unix://"):
            path = endpoint[7:]
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.sock.connect(path)
        elif endpoint.startswith("tcp://"):
            addr_port = endpoint[6:]
            tokens = addr_port.split(":")
            addr = tokens[0]
            port = int(tokens[1])
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.connect((addr, port))
        elif endpoint.startswith("ssl"):
            scheme = endpoint.split(":")[0]
            assert scheme == "ssl-tcp" or scheme == "ssl", \
                "Invalid endpoint %s" % endpoint
            addr_port = endpoint[(len(scheme) + 3):]
            tokens = addr_port.split(":")
            addr = tokens[0]
            port = int(tokens[1])
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock = ssl.SSLSocket(sock=sock, certfile=sec_ctx.certfile, keyfile=sec_ctx.keyfile,
                                      ca_certs=sec_ctx.ca_certs, ssl_version=ssl.PROTOCOL_TLSv1_2)
            self.sock.connect((addr, port))
        else:
            assert False, "Invalid endpoint %s" % endpoint

    def _OnConnect(self):
        self._StartReceiving()
        self._SendConnectionSettings(self.endpoint_ids)
        # monitor outbound connection via keepalives
        self.keep_alive_timeout_sec = KEEPALIVE_CLIENT_DEFAULT_TIMEOUT_SEC
        self._StartKeepAlive()

    def _ConnectRetry(self, endpoint, sec_ctx):
        last_exception = None
        # Acquire the lock to prevent channels from attempting to send data
        # while the socket is down
        with self.send_lock:
            for i in range(self.retry_policy.attempts):
                try:
                    self._Connect(endpoint, sec_ctx)
                    LOGGER.debug("NsxRpcConnection (%s) Successful connection "
                                 "(attempt %s)", self, i + 1)
                    return
                except (ValueError, gevent.ssl.SSLError) as e:
                    # do not retry after non-network errors (fast path)
                    LOGGER.debug(
                        "NsxRpcConnection (%s) Failed connection ", self)
                    raise e
                except (socket.error, socket.timeout) as e:
                    last_exception = e
                    LOGGER.debug("NsxRpcConnection (%s) Failed connection "
                                 "(attempt %s)", self, i + 1)
                    gevent.sleep(self.retry_policy.delay)
            else:
                raise last_exception

    def AcceptOne(self, endpoint):
        """ DEPRECATED API, only for use by NSX Container Plugin.

        Binds to endpoint, starts listening, accepts one connection, and
        closes the bound socket. When this method returns, "self" can
        be used to make RPC calls through this connection.

        Only supports TCP connections meaning that endpoint must start with
        "tcp://". To listen on any interface use "*", e.g. "tcp://*:8080"

        Exceptions thrown by bind/listen/accept (e.g. socket.error) are
        propagated up. In other words, this method does not catch such
        exceptions.
        """

        if not endpoint.startswith("tcp://"):
            raise ValueError("endpoint must start with tcp://")

        addr_port = endpoint[6:]
        tokens = addr_port.split(":")
        addr = '' if tokens[0] == "*" else tokens[0]
        port = int(tokens[1])

        self.sock = None
        self.client_addr = None
        try:
            listen_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            listen_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            listen_sock.bind((addr, port))
            listen_sock.listen(1)
            self.sock, self.client_addr = listen_sock.accept()
        finally:
            listen_sock.close()

        self._StartReceiving()
        self._SendConnectionSettings(self.endpoint_ids)
        # monitor inbound connection via keepalives
        self.keep_alive_timeout_sec = KEEPALIVE_SERVER_DEFAULT_TIMEOUT_SEC
        self._StartKeepAlive()

    def Send(self, stream_id, stream_control, frame, payload=None):
        """ Attempts to send the message composed on the given parameters.

        If self is not connected (never connected or already closed) does
        nothing. This behavior is fine since we will notice and react to
        connection failures in Recv() method, which is always running.
        """
        assert (stream_control is not None or frame is not None), \
            "StreamControl or Frame can't both be None"
        stream = BinStream()
        header = NsxRpcMsgHeader(0, len(stream_control or []),
                                 len(frame or []), len(payload or []),
                                 stream_id)
        LOGGER.debug("NsxRpcConnection (%s) Send header (%d %d %d %s)",
                     self,
                     header.stream_control_size,
                     header.frame_size,
                     header.payload_size,
                     header.stream_id)

        header.encode(stream)
        if stream_control:
            stream.write_data(stream_control)
        if frame:
            stream.write_data(frame)
        if payload:
            stream.write_data(payload)

        with self.send_lock:
            # We are checking self.sock here instead of at the top because
            # another greenlet could have set sock to None while we were
            # waiting for send_lock. Once we acquired the lock, we know that
            # no other greenlet will be running. Therefore, it is safe to
            # call sendall after checking that sock is not None
            if self.sock is None:
                LOGGER.warning("NsxRpcConnection (%s) ignoring an attempt "
                               "to send on a closed connection", self)
                return
            self.sock.sendall(stream.data)

    def _StartReceiving(self):
        self.receiver = gevent.spawn(self._Receiver)
        self.receiver.link_exception(self._OnReceiverDie)

    def _RecvAll(self, size):
        """Either receives size of data and returns it or throws
        NsxRpcCallError with status code of COMMUNICATION_ERROR.
        Will attempt to reconnect if a network error occurs while
        trying to receive data.
        Blocks until one of the above happens.
        """
        if self.sock is None:
            raise channel.NsxRpcCallError(CallStatus(
                CallStatus.COMMUNICATION_ERROR,
                "Can't receive on closed connection"))

        try:
            rest = size
            data = []
            while rest > 0:
                new_data = self.sock.recv(rest)
                if not new_data:
                    self.Close(CloseReason(CloseReason.CLOSED_BY_REMOTE))
                    raise channel.NsxRpcCallError(CallStatus(
                        CallStatus.COMMUNICATION_ERROR,
                        "Connection closed by peer"))
                data.append(new_data)
                rest -= len(new_data)
            return b''.join(data)
        except socket.error as e:
            self.Close(CloseReason(CloseReason.NETWORK_ERROR))
            raise channel.NsxRpcCallError(CallStatus(
                CallStatus.COMMUNICATION_ERROR,
                "Connection failed with %s" % e))

    def Recv(self):
        """ Blocks until it receives a complete NSX RPC message or
        encounters an error.

        Complete message is returned as a four tuple of binary strings
            (stream_id, stream_control, frame, payload)

        Error is thrown as NsxRpcCallError with status code set
        to COMMUNICATION_ERROR. The connection is closed in this case.
        """
        stream = BinStream()
        stream.data = self._RecvAll(NsxRpcMsgHeader.SIZE)
        header = NsxRpcMsgHeader.decode(stream)
        LOGGER.debug("NsxRpcConnection (%s) decoded header (%d %d %d %s)",
                     self,
                     header.stream_control_size,
                     header.frame_size,
                     header.payload_size,
                     header.stream_id)
        rest = (header.stream_control_size +
                header.frame_size + header.payload_size)
        stream.data += self._RecvAll(rest)
        return (header.stream_id,
                stream.data[:header.stream_control_size],
                stream.data[header.stream_control_size:
                            (header.stream_control_size + header.frame_size)],
                stream.data[(header.stream_control_size + header.frame_size):])

    def Close(self, reason=CloseReason(CloseReason.UNKNOWN_REASON)):
        LOGGER.debug("NsxRpcConnection (%s) Close %s", self, reason)

        if self.retry_policy is not None and reason.code == CloseReason.NETWORK_ERROR:
            self._Reconnect()
            # We only return from _Reconnect on the main thread
        else:
            self._Shutdown(reason)

    def _Shutdown(self, reason):
        if self.sock:
            self.sock.close()
            self.sock = None

        # Close all channels using this connection
        for chan in list(self.channels.values()):
            chan.Close(reason)

        # Stop all greenlets
        self._StopGreenlets(kill_reconnect=True)

    def _Reconnect(self):
        LOGGER.debug("NsxRpcConnection (%s) _Reconnect", self)

        # This function may be called by any of the three permanent greenlets
        # (main, keep_alive and receiver) so we need to make sure that only
        # one reconnect routine is spawned
        if self.reconnect_routine is None:
            self.reconnect_routine = gevent.spawn(self._ReconnectRoutine)
            self.reconnect_routine.link(self._OnReconnectRoutineDone)
            self.reconnect_routine.link_exception(self._OnReconnectRoutineDie)

        # The KA and receiver greenlets are killed by the reconnect greenlet so
        # only the main thread will resume after this
        self.reconnect_routine.join()

    def _ReconnectRoutine(self):
        LOGGER.debug("NsxRpcConnection (%s) _ReconnectRoutine greenlet", self)

        # Kill background greenlets and start them again after successful reconnection
        self._StopGreenlets(kill_reconnect=False)

        try:
            self._ConnectRetry(self.endpoint, self.sec_ctx)
            self._OnConnect()
        except Exception as e:
            LOGGER.warning("NsxRpcConnection (%s) Failed to reconnect", self)
            self.Close(CloseReason(CloseReason.LOCAL_CLOSE))
            raise e

    def _OnReconnectRoutineDone(self, *args):
        self.reconnect_routine = None

    def _OnReconnectRoutineDie(self, *args):
        LOGGER.error("NsxRpcConnection (%s) closing connection because "
                     "reconnect_routine greenlet leaked unexpected "
                     "exception: %s", self, args.exception)
        self.Close(CloseReason(CloseReason.UNKNOWN_REASON))

    def _ProcessConnectionSettings(self, settings):
        # currently not using any connection settings, just do debug log here
        LOGGER.debug("NsxRpcConnection (%s) received connection settings",
                     self)
        for endpoint_id in settings.endpoint_id:
            LOGGER.debug("NsxRpcConnection (%s) peer endpoint_id %s",
                         self, endpoint_id)

    def _ProcessKeepAliveRequest(self, req):
        LOGGER.debug("NsxRpcConnection (%s) received keepalive request "
                     "with id %d", self, req.request_id)
        response_frame = nsx_rpc_pb2.FrameMsg()
        response_frame.keep_alive_response.request_id = req.request_id
        response_frame.keep_alive_response.timestamp = req.timestamp if req.HasField('timestamp') else 0
        self.Send(uuid.UUID(int=0), None, response_frame.SerializeToString())

    def _ProcessKeepAliveResponse(self, resp):
        LOGGER.debug("NsxRpcConnection (%s) received keepalive response "
                     "with id %d", self, resp.request_id)
        if (resp.request_id > self.keep_alive_counter or
           resp.request_id <= self.keep_alive_last_recv_id):
            LOGGER.warning("NsxRpcConnection (%s) keepAlive response id is "
                           "not valid: %d (not in range %d %d)",
                           self, resp.request_id, self.keep_alive_last_recv_id,
                           self.keep_alive_counter)
            return
        if resp.request_id != self.keep_alive_last_recv_id + 1:
            # this is concerning, since we dropped or reordered messages, but
            # it is not completely wrong - log a warning for future debugging
            LOGGER.warning("NsxRpcConnection (%s) unexpected keepalive "
                           "response: got %d expected %d", self,
                           resp.request_id, self.keep_alive_last_recv_id + 1)
        self.keep_alive_last_recv_id = resp.request_id

    def _ProcessControlFrame(self, frame_bytes):
        frame = nsx_rpc_pb2.FrameMsg()
        frame.ParseFromString(frame_bytes)
        if frame.HasField("settings"):
            self._ProcessConnectionSettings(frame.settings)
        if frame.HasField("keep_alive_request"):
            self._ProcessKeepAliveRequest(frame.keep_alive_request)
        if frame.HasField("keep_alive_response"):
            self._ProcessKeepAliveResponse(frame.keep_alive_response)

    def _Receiver(self):
        try:
            LOGGER.debug("NsxRpcConnection (%s) _Receiver listening greenlet",
                         self)
            while True:
                (stream_id, stream_control, frame_bytes, payload) = self.Recv()
                # Got a packet on this connection, reset keep_alive_budget
                self.remaining_keep_alive_budget = self.keep_alive_timeout_sec
                if stream_id.int == 0:
                    self._ProcessControlFrame(frame_bytes)
                    continue
                if stream_id not in self.channels:
                    LOGGER.warning("NsxRpcConnection (%s) dropping received "
                                   "message on unknown stream", self)
                    continue
                self.channels[stream_id].Parse(stream_id, stream_control,
                                               frame_bytes, payload)
            LOGGER.debug("NsxRpcConnection (%s) _Receiver exiting normally",
                         self)
        except channel.NsxRpcCallError as e:
            LOGGER.debug("NsxRpcConnection (%s) closing connection because "
                         "of close reason %s", self, e)
            self.Close(CloseReason(CloseReason.PROTOCOL_ERROR))

    def _OnReceiverDie(self, args):
        LOGGER.error("NsxRpcConnection (%s) closing connection because "
                     "receiver greenlet leaked unexpected exception: %s",
                     self, args.exception)
        self.Close(CloseReason(CloseReason.UNKNOWN_REASON))

    def _SendKeepAlive(self):
        self.keep_alive_counter += 1
        frame = nsx_rpc_pb2.FrameMsg()
        frame.keep_alive_request.request_id = self.keep_alive_counter
        if self.sock is None:
            LOGGER.warning("NsxRpcConnection (%s) closing connection "
                           "because socket is empty", self)
            self.Close(CloseReason(CloseReason.PROTOCOL_ERROR))
            return False
        try:
            self.Send(uuid.UUID(int=0), None, frame.SerializeToString())
            return True
        except socket.error as e:
            LOGGER.warning("NsxRpcConnection (%s) closing connection "
                           "because connection failed with %s",
                           self, e)
            self.Close(CloseReason(CloseReason.NETWORK_ERROR))
            return False

    def _KeepAliveRoutine(self):
        LOGGER.debug("NsxRpcConnection (%s) _KeepAliveRoutine greenlet", self)
        keep_alive_interval = (self.keep_alive_timeout_sec /
                               self.keep_alive_num_pings)
        self.remaining_keep_alive_budget = self.keep_alive_timeout_sec
        while True:
            # if we have not received ANY keepalive, we do not know if
            # the other endpoint is keepalive-aware
            # TODO: remove this code once we bump the protocol version
            if (self.keep_alive_last_recv_id == 0):
                self.remaining_keep_alive_budget = self.keep_alive_timeout_sec
                LOGGER.debug("NsxRpcConnection (%s) _KeepAliveRoutine no keep_alive", self)

            if self.remaining_keep_alive_budget == self.keep_alive_timeout_sec:
                LOGGER.debug("NsxRpcConnection (%s) _KeepAliveRoutine known timeout", self)
                # Budget was reset to full. This means packet received in this
                # connection during last budget period.
                # Therefore, no need to send keep_alive request
                # Sleep for keep_alive_interval/2 seconds
                keep_alive_sleep_sec = keep_alive_interval / 2
                self.remaining_keep_alive_budget -= keep_alive_sleep_sec
            else:
                # next sleep interval calculation
                keep_alive_sleep_sec = min(keep_alive_interval,
                                           self.remaining_keep_alive_budget)
                self.remaining_keep_alive_budget -= keep_alive_sleep_sec
                LOGGER.debug("NsxRpcConnection (%s) keep_alive_period_sec"
                             " %.2f remaining_keep_alive_budget %.2f",
                             self, keep_alive_sleep_sec,
                             self.remaining_keep_alive_budget)
                # 0 means we ran out of time while waiting for keepalives
                if keep_alive_sleep_sec == 0:
                    LOGGER.warning("NsxRpcConnection (%s) closing connection "
                                   "because Keepalive expired: timeout %dsec,"
                                   " last ping %d, last pong %d",
                                   self, self.keep_alive_timeout_sec,
                                   self.keep_alive_counter,
                                   self.keep_alive_last_recv_id)
                    self.Close(CloseReason(CloseReason.KEEPALIVE_EXPIRED))
                    return

                # Send keep alive request
                if self._SendKeepAlive() is not True:
                    return
            gevent.sleep(keep_alive_sleep_sec)
        LOGGER.debug("NsxRpcConnection (%s) _KeepAliveRoutine exiting", self)

    def _OnKeepAliveRoutineDie(self, args):
        LOGGER.error("NsxRpcConnection (%s) closing connection because "
                     "keep_alive_routine greenlet leaked unexpected "
                     "exception: %s",
                     self, args.exception)
        self.Close(CloseReason(CloseReason.UNKNOWN_REASON))

    def _SendConnectionSettings(self, endpoint_ids):
        LOGGER.debug("NsxRpcConnection (%s) _SendConnectionSettings", self)
        frame = nsx_rpc_pb2.FrameMsg()
        frame.settings.version = 1
        for endpoint_id in endpoint_ids:
            frame.settings.endpoint_id.append(endpoint_id)
        self.Send(uuid.UUID(int=0), None, frame.SerializeToString())

    def _StartKeepAlive(self):
        assert(self.keep_alive_timeout_sec > 0)
        LOGGER.debug("NsxRpcConnection (%s) _StartKeepAlive timeout %d",
                     self, self.keep_alive_timeout_sec)
        self.keep_alive_routine = gevent.spawn(self._KeepAliveRoutine)
        self.keep_alive_routine.link_exception(self._OnKeepAliveRoutineDie)

    def _StopGreenlets(self, kill_reconnect=False):
        LOGGER.debug("NsxRpcConnection (%s) _StopGreenlets ", self)
        if self.keep_alive_routine:
            # terminate the KA greenlet
            if not self.keep_alive_routine.dead:
                LOGGER.debug("NsxRpcConnection (%s) _StopGreenlets: KA not dead, killing it ", self)
                self.keep_alive_routine.kill()
            self.keep_alive_routine = None
        if self.receiver:
            # terminate the receive greenlet
            if not self.receiver.dead:
                LOGGER.debug("NsxRpcConnection (%s) _StopGreenlets: Receiver not dead, killing it ", self)
                self.receiver.kill()
            self.receiver = None
        if self.reconnect_routine and kill_reconnect:
            # terminate the reconnect greenlet
            if not self.reconnect_routine.dead:
                LOGGER.debug("NsxRpcConnection (%s) _StopGreenlets: Reconnect not dead, killing it ", self)
                self.reconnect_routine.kill()
            self.reconnect_routine = None
